function [lower,upper,pi_lower,pi_upper,f]...
    = r17_parameters(para,Y,D,Z,opts,opts_val )
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% DESCRIPTION: This function computes bounds on functionals of the joint
% distribution of potential outcomes as described in Russell (2017).
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% INPUTS:
%
%   parameter_type : the parameter the researcher wishes to bound:
%            (i) 'jp' = joint probability of (yo,y1) for all values of y.
%           (ii)'ate' = average treatment effect
%          (iii) 'vc' = voting criterion (i.e. P(y1>y0))
%           (iv) 'cp' = P(yd=y| y(d-1) = y') for all values of y,y'.
%            (v) 'mp' = marginal probability P(yd =y) for all y,d.
%           (vi) 'en' = entropy of the joint distribution
%          (vii)'cor' = correlation of yo,y1.
%         (viii) 'v0' = variance of y0.
%           (ix) 'v1' = variance of y1.
%            (x)'v10' = variance of y0 and y1
%
%   Y : n x 1 outcome variable data vector.
%
%   D : n x 1 treatment variable data vector.
%
%   Z : n x 1 instrument data vector (set empty if no instruments)
%
%   options : desired options (i.e. either 'none' or 'MTR')
%
%   options_value : desired MTR level
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% OUTPUTS:
%
%   lower : the lower bound on the specified parameter
%
%   upper : the upper bound on the specified parameter
%
%   pi_lower : the distribution that obtained the lower bound on the
%             specified parameter
%
%   pi_upper : the distribution that obtained the upper bound on the
%             specified parameter
%
%   f :  for linear parameters, this is a vector that determines the
%       objective function (e.g. obj.func.=f'pi).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%Calculates estimate of empirical distribution P(Y=i,D=j|Z=k)
[ p, ~ , ~,~,~] = r17_empdist(Y,D,Z);

%Compute median of observed value of Y, used for computing marginal and
%joint probabilities of Y0 and Y1 and conditional probabilities of
%P(Y1|Y0)
medy=median(Y);

Y=unique(Y,'rows');
D=unique(D,'rows');
J=size(p,2);
M=length(Y);
K=length(D);

if K>2
    error('This program calculates bounds only for binary treatment');
end

%In case Y is not integer, reassign integer values
Ytest=(1:M)';

cu=M^K; %# of nodes in the U part
cy=K*M; %# of nodes in the Y part

%Generate a matrix with all posible combinations of (Y,D)
comb_y=combinate_vec(Ytest,D);

%Generate a matrix Gp that records which pair (y,d) is possible for each
%(y0=i,y1=j,d) for all i,j.

Gp=NaN(2*cu,cy);

iter=1;

for i=1:M                   %Y0 loop
    for j=1:M               %Y1 loop
        for d=0:1           %D  loop 
            Gp(iter,:)=ismember(comb_y,[d*j+(1-d)*i,d],'rows')';
            iter=iter+1;
        end
    end
end

Gp=Gp';

%NOTE: the mapping between observed distribution "p" and joint distribution
%"pi" is given by: p=Gp*pi.

%**************************************************************************
%Example: for Y,D in {0,1}, let pij=P(Y=i,D=j), pi_ijd=P(Y0=i,Y1=j,D=d)
%
% Then: 
%           p =  Gp * pi
%
% is equivalent to:
%          |p00|    =   |  1 0 1 0 0 0 0 0 | |pi_000|
%          |p01|        |  0 1 0 0 0 1 0 0 | |pi_001|
%          |p10|        |  0 0 0 0 1 0 1 0 | |pi_010|
%          |p11|        |  0 0 0 1 0 0 0 1 | |pi_011|
%                                            |pi_100|
%                                            |pi_101|
%                                            |pi_110|
%                                            |pi_111|
%
%Note: all parameters ordered in lexicographic order!
%**************************************************************************

%Matrix used to group values of the joint pi(Y0,Y1,D) to recover the
%distribution Q(Y0,Y1). The mapping between unobserved distribution "Q" and 
%joint distribution "pi" is given by: Q=Gq*pi.
Gq=kron(eye(cu),[1,1]);

%Take the Kronecker product with eye(J) to impose constraints for every
%value of the instrument.
Gp=kron(eye(J),Gp);

%Review MTR option input:
if strcmp(opts,'MTR')
    
    comb_u=permn(unique(Y),K);
    
    if opts_val<=0
        
        A_opt_con=(comb_u(:,2)>comb_u(:,1))'+0;
        b_opt_con=-opts_val;
        
    else
        
        A_opt_con=-(comb_u(:,2)>comb_u(:,1))'+0;
        b_opt_con=-opts_val;
        
    end
    
    A=kron(eye(J),A_opt_con*Gq);
    b=repmat(b_opt_con,[J,1]);
    
elseif strcmp(opts,'none') || strcmp(opts,'None') || strcmp(opts,'NONE')
    
    A=[];
    b=[];
else
    
    error('Please Enter a Valid Option')
    
end

%Review parameter input:
if strcmp(para,'jp')%Bounds on the joint distribution
    
    [ lower,upper,pi_lower,pi_upper,f] = r17_jp( p,Y,D,Gp,Gq,...
        A,b,medy);
    
elseif strcmp(para,'ate')%Bounds on the average treatment effect
    
    [bounds_ate,pi_lower,pi_upper,f] = r17_ate( p,Y,D , Gp,Gq,...
        A,b);
    
    lower=bounds_ate(1,1);
    upper=bounds_ate(1,2);
    
elseif strcmp(para,'vc')%Bounds on the voting criterion
    
    [ bounds_vc,pi_lower,pi_upper,f] = r17_vc( p,Y,D , Gp,Gq,...
        A,b);
    
    lower=bounds_vc(1,1);
    upper=bounds_vc(1,2);
    
elseif strcmp(para,'cp')%Bounds on the conditional distribution
    
    [lower,upper,pi_lower,pi_upper,f] = r17_cp( p,Y,D, Gp,Gq,...
        A,b,medy);
    
elseif strcmp(para,'mp')%Bounds on the marginal distribution
    
    [lower,upper,pi_lower,pi_upper,f ] = r17_mp( p,Y,D,Gp,Gq,...
        A,b,medy);
    
elseif strcmp(para,'corr')%Bounds on the correlation
    
    [lower,upper,pi_lower,pi_upper] = r17_corr( p,Y,D, Gp,Gq,...
        A,b);
    
    f=[];
    
elseif strcmp(para,'var10')%Bounds on the unconditional variance of Y1-Y0
    
    [lower,upper,pi_lower,pi_upper] = r17_var10( p,Y,D, Gp,Gq,...
        A,b);
    
    f=[];
    
else
    error('Invalid Parameter Selected');
end

end

function [lowb,uppb,pi_lower,pi_upper,fmat] = r17_jp(p,Y,D,Gp,Gq,...
    A,b,mid)

K=length(unique(D));
J=size(p,2);
q_legend=permn(sort(unique(Y,'rows')),K);
mid1=mid;
mid2=mid;

Aeq=Gp;
beq=p(:);

for j=1:J-1
    tmp=zeros(1,J);
    tmp(1,j)=1;
    tmp(1,j+1)=-1;
    Aeq=[Aeq;kron(tmp,Gq)];
    beq=[beq;zeros(size(Gq,1),1)];
end

lb=zeros(size(Gp,2),1);
ub=ones(size(Gp,2),1);

f=(q_legend(:,1)<=mid1 & q_legend(:,2)>mid2);
f=repmat(Gq,[1,J])'*f;

[pi_lower,lowb,~] = gurobilinprog(f, A, b, Aeq, beq, lb, ub);
[pi_upper,uppb,~] = gurobilinprog(-f, A, b, Aeq, beq, lb, ub);

lowb=lowb/J;
uppb=-uppb/J;

fmat=[];

end

function [bounds,pi_lower,pi_upper,f] = r17_ate(p,Y,D,Gp,Gq,...
    A,b)

K=length(unique(D));
J=size(p,2);

comb_u=permn(unique(Y),K);

f=repmat(Gq,[1,J])'*(comb_u(:,2)-comb_u(:,1));

Aeq=Gp;
beq=p(:);

for j=1:J-1
    tmp=zeros(1,J);
    tmp(1,j)=1;
    tmp(1,j+1)=-1;
    Aeq=[Aeq;kron(tmp,Gq)];
    beq=[beq;zeros(size(Gq,1),1)];
end

lb=zeros(size(Aeq,2),1);
ub=ones(size(Aeq,2),1);

%Output is lower bound (lowb) and joint distribution of (Y0,Y1,D) that
%obtains the lower bound (pi_lower)
[pi_lower,lowb,~] = gurobilinprog(f, A, b, Aeq, beq, lb, ub);

%Output is upper bound (uppb) and joint distribution of (Y0,Y1,D) that
%obtains the lower bound (pi_upper)
[pi_upper,uppb,~] = gurobilinprog(-f, A, b, Aeq, beq, lb, ub);

bounds=[lowb,-uppb]/J;

end

function [ bounds, pi_lower,pi_upper,f ] = r17_vc( p,Y,D,Gp,Gq,...
    A,b)

K=length(unique(D));
M=length(unique(Y));
J=size(p,2);

comb_u=permn(unique(Y),K);
f=(comb_u(:,2)>comb_u(:,1))+0;
f=repmat(Gq,[1,J])'*f;

Aeq=Gp;
beq=p(:);

for j=1:J-1
    tmp=zeros(1,J);
    tmp(1,j)=1;
    tmp(1,j+1)=-1;
    Aeq=[Aeq;kron(tmp,Gq)];
    beq=[beq;zeros(size(Gq,1),1)];
end

lb=zeros(size(Gp,2),1);
ub=ones(size(Gp,2),1);

[pi_lower,lowb,~] = gurobilinprog(f, A, b, Aeq, beq, lb, ub);
[pi_upper,uppb,~] = gurobilinprog(-f, A, b, Aeq, beq, lb, ub);

bounds=[lowb,-uppb]/J;

end

function [ lowb, uppb, pi_lower,pi_upper,f] = r17_cp(p,Y,D,Gp,Gq,...
    A,b,mid)

K=length(unique(D));
M=length(unique(Y));
J=size(p,2);

q_legend=permn(sort(unique(Y,'rows')),K);

d1=kron(eye(J),Gq'*(q_legend(:,1)<=mid));
d2=kron(eye(J),ones(M^K*K,1));

A=[A,zeros(size(A,1),1)];

Aeq=[Gp , -p(:);
    d1',  zeros(J,1)   ;
    d2', -ones(J,1)  ];
beq=[zeros(length(p(:)),1);ones(J,1);zeros(J,1)];

for j=1:J-1
    tmp=zeros(1,J);
    tmp(1,j)=1;
    tmp(1,j+1)=-1;
    tmp1=kron(tmp,Gq);
    tmp2=[tmp1,zeros(size(tmp1,1),1)];
    Aeq=[Aeq;tmp2];
    beq=[beq;zeros(size(Gq,1),1)];
end

lb=[repmat(Gq,[1,J])'*zeros(M^K,1);1];
ub=[repmat(Gq,[1,J])'*ones(M^K,1);Inf];

f=(q_legend(:,1)<=mid & q_legend(:,2)>mid);
f=[repmat(Gq,[1,J])'*f;0];

[pi_lower,lowb,~] = gurobilinprog(f, A, b, Aeq, beq, lb, ub);
[pi_upper,uppb,~] = gurobilinprog(-f, A, b, Aeq, beq, lb, ub);

lowb=lowb/J;
uppb=-uppb/J;

end

function [ lower_mat, upper_mat,pi_lower,pi_upper,fmat] = r17_mp(p,Y,D,...
    Gp,Gq,A,b,mid)
K=length(unique(D));
J=size(p,2);

Aeq=Gp;
beq=p(:);

for j=1:J-1
    tmp=zeros(1,J);
    tmp(1,j)=1;
    tmp(1,j+1)=-1;
    Aeq=[Aeq;kron(tmp,Gq)];
    beq=[beq;zeros(size(Gq,1),1)];
end

lb=zeros(size(Gp,2),1);
ub=ones(size(Gp,2),1);

comb_y=permn(unique(Y),K);

lower_mat=NaN(1,2);
upper_mat=NaN(1,2);

fmat=NaN(size(Aeq,2),2);

pi_lower=NaN(size(Aeq,2),2);
pi_upper=NaN(size(Aeq,2),2);

f=repmat(Gq,[1,J])'*(comb_y(:,1)<=mid)+0;

[pi_lower(:,1),lower_mat(:,1),~] = gurobilinprog(f,A,b,Aeq,beq,lb,ub);
[pi_upper(:,1),upper_mat(:,1),~] = gurobilinprog(-f,A,b,Aeq,beq,lb,ub);

f=repmat(Gq,[1,J])'*(comb_y(:,2)>mid)+0;

[pi_lower(:,2),lower_mat(:,2),~] = gurobilinprog(f,A,b,Aeq,beq,lb,ub);
[pi_upper(:,2),upper_mat(:,2),~] = gurobilinprog(-f,A,b,Aeq,beq,lb,ub);

lower_mat=lower_mat/J;
upper_mat=-upper_mat/J;

end

function [ lowb,uppb, pi_lower,pi_upper] = r17_corr(p,Y,D,Gp,Gq,...
    A,b)

K=length(unique(D));
M=length(unique(Y));
J=size(p,2);

Aeq=Gp;
beq=p(:);

for j=1:J-1
    tmp=zeros(1,J);
    tmp(1,j)=1;
    tmp(1,j+1)=-1;
    Aeq=[Aeq;kron(tmp,Gq)];
    beq=[beq;zeros(size(Gq,1),1)];
end

q_legend1=permn(sort(unique(Y,'rows')),K);
q_legend2=permn(1:M,K);

opt = optimset('GradObj','on','MaxIter',10^8, ...
    'TolX',1e-15,'TolFun',1e-8,'TolCon',1e-8,'UseParallel',true);

lb=zeros(size(Gp,2),1);
ub=ones(size(Gp,2),1);

%Use linear program to determine initial point
f=repmat(Gq,[1,J])'*ones(M^K,1);
[theta0,~,~] = gurobilinprog(f, A, b, Aeq, beq, lb, ub);

lb=zeros(size(Gp,2),1)-10^-6;
ub=ones(size(Gp,2),1);

[pi_lower,lowb]...
    = knitromatlab(@(theta)corr_l(theta,Gq,J,q_legend1,q_legend2),theta0,...
    A,b,Aeq,beq,lb,ub,[],[],opt,[]);

[pi_upper,uppb]...
    = knitromatlab(@(theta)corr_u(theta,Gq,J,q_legend1,q_legend2),theta0,...
    A,b,Aeq,beq,lb,ub,[],[],opt,[]);

lowb=lowb/J;
uppb=-uppb/J;

end

function [val,gradf] = corr_l(pii,Gq,J,q_legend1,q_legend2)

q=Gq*pii(1:(end/J),1);

avg=q'*q_legend1;

num=q'*((q_legend1(:,1)-avg(1,1)).*(q_legend1(:,2)-avg(1,2)));

margy0=accumarray(q_legend2(:,1),q);
margy1=accumarray(q_legend2(:,2),q);

y0=unique(q_legend1(:,1),'rows');
y1=unique(q_legend1(:,2),'rows');

denom=sqrt(margy0'*(y0-avg(1,1)).^2)*sqrt(margy1'*(y1-avg(1,2)).^2);

val=num/denom;

%%%%%% Compute Derivative of Correlation %%%%%%

%%%Derivative Broken into Numerator and Denominator, and sections
%%%based on the product rule...:)

d_num1=(q_legend1(:,1)-avg(1,1)).*(q_legend1(:,2)-avg(1,2));
d_num2=sum(q.*(-q.*q_legend1(:,1)).*(q_legend1(:,2)-avg(1,2)));
d_num3=sum(q.*(q_legend1(:,1)-avg(1,1)).*sum(-q.*q_legend1(:,2)));

d_num=d_num1+d_num2+d_num3;

d_denom1=(margy0'*(y0-avg(1,1)).^2);
d_denom2=((q_legend1(:,1)-avg(1,1)).^2)+...
    2*sum(q.*(-q.*q_legend1(:,1)).*(q_legend1(:,1)-avg(1,1)));
d_denom3=sqrt(margy1'*(y1-avg(1,2)).^2);
d_denom4=(margy1'*(y1-avg(1,2)).^2);
d_denom5=((q_legend1(:,2)-avg(1,2)).^2)+...
    2*sum(q.*(-q.*q_legend1(:,2)).*(q_legend1(:,2)-avg(1,2)));
d_denom6=sqrt(margy0'*(y0-avg(1,1)).^2);

d_denom=0.5*(d_denom1.^-0.5).*d_denom2.*d_denom3+...
    0.5*(d_denom4.^-0.5).*d_denom5.*d_denom6;

gradf=(denom.*d_num-num.*d_denom)./(denom.^2);

gradf=repmat(Gq,[1,J])'*gradf;

end

function [val,gradf] = corr_u(pii,Gq,J,q_legend1,q_legend2)

[val,gradf] = corr_l(pii,Gq,J,q_legend1,q_legend2);

val=-val;

gradf=-gradf;

end

function [ lowb,uppb, pi_lower,pi_upper] = r17_var10(p,Y,D,Gp,Gq,...
    A,b)

K=length(unique(D));
M=length(unique(Y));
J=size(p,2);

Aeq=Gp;
beq=p(:);

for j=1:J-1
    tmp=zeros(1,J);
    tmp(1,j)=1;
    tmp(1,j+1)=-1;
    Aeq=[Aeq;kron(tmp,Gq)];
    beq=[beq;zeros(size(Gq,1),1)];
end

q_legend1=permn(sort(unique(Y,'rows')),K);

opt = optimset('GradObj','on','MaxIter',10^8, ...
    'TolX',1e-15,'TolFun',1e-8,'TolCon',1e-8,'UseParallel',true);

lb=zeros(size(Gp,2),1);
ub=ones(size(Gp,2),1);

%Use linear program to determine initial point
f=repmat(Gq,[1,J])'*ones(M^K,1);
[theta0,~,~] = gurobilinprog(f, A, b, Aeq, beq, lb, ub);

lb=zeros(size(Gp,2),1)-10^-6;
ub=ones(size(Gp,2),1);

[pi_lower,lowb,~]...
    = knitromatlab(@(theta)var10_l(theta,Gq,J,q_legend1),theta0,...
    A,b,Aeq,beq,lb,ub,[],[],opt,[]);

[pi_upper,uppb,~]...
    =  knitromatlab(@(theta)var10_u(theta,Gq,J,q_legend1),theta0,...
    A,b,Aeq,beq,lb,ub,[],[],opt,[]);

lowb=lowb/J;
uppb=-uppb/J;

end

function [val,gradf] = var10_l(pii,Gq,J,q_legend1)

q=Gq*pii(1:(end/J),1);

y=q_legend1(:,2)-q_legend1(:,1);

avg=q'*y;

val=q'*(y-avg).^2;

gradf=((y-avg).^2)+2*sum(q.*(-q.*y).*(y-avg));

gradf=repmat(Gq,[1,J])'*gradf;

end

function [val,gradf] = var10_u(pii,Gq,J,q_legend1)

[val,gradf] = var10_l(pii,Gq,J,q_legend1);

val=-val;

gradf=-gradf;
end
